import sys

# Global imports when running the modules for testing
if "-m" not in sys.argv:

    # Import the functions so that they are accesible from the main module
    from .labeling.segmentlabeler import SegmentLabeler  # noqa
    from .labeling.volumelabeler import VolumeLabeler  # noqa



from matplotlib.backend_bases import MouseButton
from matplotlib.colors import BoundaryNorm
import matplotlib.pyplot as plt
import numpy as np
from skimage.segmentation import mark_boundaries
import sys
import time
import logging


def array_in_arraylist(array, arraylist):
    for _i,  _array in enumerate(arraylist):
        if np.array_equal(array, _array):
            return True
    return False


def unique_arrays(lst: list):
    used = []
    _ = [used.append(x) for x in lst if not array_in_arraylist(x, used)]
    return used


class SegmentLabeler:

    def __init__(self,
                 img,
                 segmentation,
                 labeldict={'moho': 1, '410': 2, '660': 3, 'none': 9999},
                 loglevel=logging.INFO):
        """This class takes in an image, some corresponding 
        segmentation e.g. (SLIC) and a label dictionary that can be used to
        label the image in an 'intuitive' GUI.

        Examples
        --------

        The usage of this labeling is fairly straight forward. Given an image,
        ``img``, and a segmentation, ``segs``, of said image, we instantiate 
        the class and call the start labeling method.

        >>> from lwsspy.ml import SegmentLabeler
        >>> sl = SegmentLabeler(img, segments)
        >>> labeled_mask = sl.start_labeling()

        This will open two figures. One contains the image by itself and the 
        second one contains the image and the outlines of the mask. We will 
        use the firs image for reference and the second image to actually 
        label the image segments. The GUI has a few controls that can be used.

        ================= ======================================
        Control           Action
        ================= ======================================
        Mouse-left        Add label to segment
        Mouse-right       Remove label to segment
        Mouse-left-drag   Add labels to segments dragged over
        Mouse-right-drag  Remove labels to segments dragged over
        n                 Next label
        p                 Previous label  True
        d                 Delete previously labeled segment
        esc               Close figure and return the currently 
                          selected mask
        ================= ======================================

        The currently selected mask will also be returned if any of the figures 
        is closed.

        The selected mask can then be viewed via 

        >>> import matplotlib.pyplot as plt
        >>> imshow(labeled_mask, aspect='auto')

        Note that depending on the values you chose in the label dictionary
        you will have to create a colormap and norm that resembles the mask 
        values.


        Parameters
        ----------
        img : ndarray [w x h x 3]
            Image
        segmentation : ndarray [w x h]
            mask that has a unique nunmber for each segment such that it can be
            labeled.
        labeldict : dict, optional
            Dictionary of labels, must contain the 'none' keyword, which 
            denotes the unlabeled value, 
            by default {'moho': 1, '410': 2, '660': 3, 'none': 9999}
        loglevel : logging.LOGLEVEL, optional
            loglevel, used to debug the event loop. Not necessary to be 
            modified, by default logging.INFO


        Notes
        -----

        :Authors:
            Lucas Sawade (lsawade@princeton.edu)

        :Last Modified:
            2021.07.02 00.00 (Lucas Sawade)

        """

        self.img = img
        self.segmentation = segmentation
        self.labeldict = labeldict

        self.safetyfirst()

        self.labeled = labeldict['none'] * np.ones_like(self.segmentation)

        # Pick variables
        self.pickhistory = dict()
        self.picklabels = []
        self.picknumber = []
        for label, number in self.labeldict.items():
            if label != 'none':
                self.pickhistory[label] = []
                self.picklabels.append(label)
                self.picknumber.append(number)
        self.activelabel = 0
        self.nlabels = len(self.picklabels)
        self.mouse_pressed = False

        # Logging
        self.loglevel = loglevel
        self.__setup_logger__()

    def safetyfirst(self):

        if 'none' not in self.labeldict:
            raise ValueError(
                "The label dictionary has to contain 'none' keywords.")

    def __setup_logger__(self):

        # create logger
        self.logger = logging.getLogger('SegmentationLabeler')
        self.logger.setLevel(self.loglevel)

        # create console handler and set level to debug
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)

        # create formatter
        formatter = logging.Formatter(
            '[%(asctime)s] %(name)s | %(levelname)8s: '
            '%(message)s (%(filename)s:%(lineno)d)',
            datefmt='%m/%d/%Y %I:%M:%S %p')

        # add formatter to ch
        ch.setFormatter(formatter)

        # add ch to logger
        if len(self.logger.handlers) > 0:
            self.logger.handlers = []

        self.logger.addHandler(ch)

        # 'application' code
        self.logger.debug('debug message')
        self.logger.info('info message')
        self.logger.warning('warn message')
        self.logger.error('error message')
        self.logger.critical('critical message')

    def plot_figure(self):

        self.logger.debug(f'Plotting Image & Segmentation...')
        self.static_figure = plt.figure(figsize=(6, 6))

        self.static_ax = plt.subplot(111)
        self.static_ax.imshow(self.img, aspect='auto')

        self.segment_figure = plt.figure(figsize=(6, 6))
        self.segment_ax = plt.subplot(111)
        self.segment_ax.imshow(
            mark_boundaries(self.img, self.segmentation, color=(0, 0, 0)),
            aspect='auto')

        self.plot_labeled_image()

    def plot_labeled_image(self):

        self.logger.debug(f'Plotting Labeled Image ...')

        # Get boundary color norm based on label numbers
        pickarray = np.array(self.picknumber)
        dpickarray = np.diff(pickarray)/2

        # Very artificial create color bounds
        self.bounds = list(pickarray[:-1] + dpickarray)
        self.bounds = [pickarray[0] - dpickarray[0]] + self.bounds
        self.bounds = self.bounds + [pickarray[-1] + dpickarray[-1]]

        # Create adhoc cmap and norm
        self.cmap = plt.get_cmap('rainbow').copy()
        self.cmap.set_bad('lightgray', alpha=0.0)

        self.norm = BoundaryNorm(self.bounds, self.cmap.N)

        # Get mask
        self.__update_labeled_image__()

    def start_labeling(self):

        self.plot_figure()
        self.segment_ax.set_title(
            f"Picking {self.picklabels[self.activelabel]}")
        plt.draw()
        self.segment_figure.canvas.mpl_connect('key_press_event', self.onkey)
        self.segment_figure.canvas.mpl_connect(
            'button_press_event', self.onclick)
        self.segment_figure.canvas.mpl_connect(
            'button_release_event', self.onrelease)
        self.segment_figure.canvas.mpl_connect(
            'motion_notify_event', self.onmotion)

        # Closing connections
        self.cidsegment = self.segment_figure.canvas.mpl_connect(
            'close_event', self.onclose_segment)
        self.cidstatic = self.static_figure.canvas.mpl_connect(
            'close_event', self.onclose_static)

        plt.show(block=True)

        return self.labeled

    def onclick(self, event):
        self.logger.debug(f'button press -- {event.button}')
        self.clicktime = time.time()
        self.mouse_pressed = True
        self.coords = []

    def onmotion(self, event):

        # duration = time.time() - self.clicktime

        # Second statement means nothing in toolbar selected
        if self.mouse_pressed \
                and self.segment_figure.canvas.manager.toolbar.mode == '':
            # and np.isclose((duration * 100) % 5, 0):
            # Get pixel locations
            x, y = int(event.xdata), int(event.ydata)
            self.logger.debug(f"Pick location:  ({x}, {y})")

            # Append coordinates
            self.coords.append((x, y))

    def onrelease(self, event):

        self.mouse_pressed = False
        self.logger.debug(f'button_release_event -- {event.button}')

        # Return if pick not in axes
        if event.inaxes != self.segment_ax:
            return

        # Return if pick  is happening while toolbar active
        state = self.segment_figure.canvas.manager.toolbar.mode
        if state == '':
            self.logger.debug(f"No figure tool selected.")
        else:
            self.logger.debug(f"{state} selected.")
            return

        duration = time.time() - self.clicktime

        # Get pixel locations
        x, y = int(event.xdata), int(event.ydata)

        self.logger.debug(f"Pick location:  ({x}, {y})")

        # Label single coordinate
        if event.button is MouseButton.LEFT and duration <= 0.3:
            self.logger.debug("Left Mouse button pressed")
            self.__label_segment__(x, y)

        # Label many coordinates
        elif event.button is MouseButton.LEFT and duration > 0.3:
            self.logger.debug("Left Mouse button dragged")
            for x, y in self.coords:
                self.__update_labeled__(y, x)
            self.__update_labeled_image__()

        # Remove single coordinates
        elif event.button is MouseButton.RIGHT and duration <= 0.3:
            self.logger.debug("Right Mouse button pressed")
            self.__remove_segment__(x, y)

        # Remove many coordinates
        elif event.button is MouseButton.RIGHT and duration > 0.3:
            self.logger.debug("Right Mouse button dragged")
            for x, y in self.coords:
                self.__reset_segment__(x, y)
            self.__update_labeled_image__()

        # Remove duplicates
        self.pickhistory[self.picklabels[self.activelabel]] = \
            unique_arrays(
                self.pickhistory[self.picklabels[self.activelabel]])

    def __label_segment__(self, x, y):

        # Update labeled and then update the image plotted on top
        self.__update_labeled__(y, x)
        self.__update_labeled_image__()

    def __remove_segment__(self, x, y):

        #  Update labeled and the update image
        self.__reset_segment__(x, y)
        self.__update_labeled_image__()

    def __reset_segment__(self, x, y):

        # Get value of segmented image
        val = self.segmentation[y, x]

        # Find where everything is that is in the segment
        pos = np.where(np.isclose(self.segmentation, val))

        # Remove the entry from the history
        for _i,  _histpos in \
                enumerate(self.pickhistory[self.picklabels[self.activelabel]]):

            if np.array_equal(_histpos, pos):
                self.pickhistory[self.picklabels[self.activelabel]].pop(_i)
                break

        # Reset the labeled array
        self.labeled[pos] = self.labeldict['none']

    def __update_labeled__(self, x, y):

        # Get value of segmented image
        val = self.segmentation[x, y]

        # Find where everything is that is in the segment
        pos = np.where(np.isclose(self.segmentation, val))

        # Add pos to history
        self.pickhistory[self.picklabels[self.activelabel]].append(pos)

        # Set Labeled to active label number
        self.labeled[pos] = self.picknumber[self.activelabel]

    def onkey(self, event):
        self.logger.debug(f'key press -- {event.key}')

        if event.key == 'n':
            self.__next_label__()

        elif event.key == 'p':
            self.__previous_label__()

        elif event.key == 'd':
            self.__remove_previous__()

        elif event.key == 'escape':
            self.__stop_labeling__()

    def __next_label__(self):

        self.activelabel += 1

        if self.activelabel == self.nlabels:
            self.activelabel -= 1

        self.segment_ax.set_title(
            f"Picking {self.picklabels[self.activelabel].capitalize()}")

        self.segment_figure.canvas.draw()

    def __previous_label__(self):

        self.activelabel -= 1

        if self.activelabel == -1:
            self.activelabel += 1

        self.segment_ax.set_title(
            f"Picking {self.picklabels[self.activelabel]}")

        self.segment_figure.canvas.draw()

    def __remove_previous__(self):
        self.logger.debug(
            f"Remove previous selection for label -- "
            f"{self.picklabels[self.activelabel]}")

        if len(self.pickhistory[self.picklabels[self.activelabel]]) == 0:
            return

        # Get last position
        pos = self.pickhistory[self.picklabels[self.activelabel]][-1]

        # Reset labeled to the  masked value
        self.labeled[pos] = self.labeldict['none']

        # Pop latest one
        self.pickhistory[self.picklabels[self.activelabel]].pop(-1)

        # Update the  labeled image
        self.__update_labeled_image__()

    def __update_labeled_image__(self):

        self.logger.debug(f'Updating Labeled Image ...')

        self.labeled_m = np.ma.masked_values(
            self.labeled, self.labeldict['none']
        )

        if hasattr(self, 'labeled_img'):
            self.labeled_img.set_data(self.labeled_m)
        else:
            self.labeled_img = self.segment_ax.imshow(
                self.labeled_m, cmap=self.cmap, norm=self.norm, aspect='auto',
                alpha=0.5
            )

        self.segment_figure.canvas.draw()

    def onclose_segment(self, event):
        self.logger.debug("Closing Segmentation Figure")

        self.static_figure.canvas.mpl_disconnect(self.cidstatic)

        self.__stop_labeling__()

    def onclose_static(self, event):
        self.logger.debug("Closing Static Figure")

        self.segment_figure.canvas.mpl_disconnect(self.cidsegment)

        self.__stop_labeling__()

    def __stop_labeling__(self):
        self.logger.debug("Stop Labeling & quitting the program")

        fignums = plt.get_fignums()
        if len(fignums) != 0:
            for num in fignums:
                plt.close(num)

        # Removed the plotted image so that it can be plotted again
        if hasattr(self, 'labeled_img'):
            del self.labeled_img

        # Stop logging



""" :noindex:
Setup.py file that governs the installatino process of
`how_to_make_a_python_package` it is used by
`conda install -f environment.yml` which will install the package in an
environment specified in that file.

"""
from setuptools import setup
from setuptools.command.test import test as testcommand

# This installs the pytest command. Meaning that you can simply type pytest
# anywhere and "pytest" will look for all available tests in the current
# directory and subdirectories recursively (not yet supported in the setup.cfg)


class PyTest(testcommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.tests")]

    def initialize_options(self):
        testcommand.initialize_options(self)
        self.pytest_args = []

    def run_tests(self):
        import pytest
        import sys
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(cmdclass={'tests': PyTest})





# External
from typing import Optional
import numpy as np
from matplotlib import colors
import matplotlib.pyplot as plt
from skimage.segmentation import slic

# Internal
from .segmentlabeler import SegmentLabeler


class MidpointNormalize(colors.Normalize):
    def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
        self.midpoint = midpoint
        colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        # I'm ignoring masked values and all kinds of edge cases to make a
        # simple example...
        x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y))


class VolumeLabeler:

    def __init__(
            self, x, y, z, V,
            labeldict={'moho': 1, '410': 2, '660': 3, 'none': 9999},
            labeled: Optional[dict] = None):
        """This class uses the 
        :class:``lwsspy.ml.labeling.segementlabeler.SegmentLabeler`` to label
        slices in a volume. After instantiation, use the label 
        method to label a Volumen of single Channel data.

        Parameters
        ----------
        x : arraylike
            x vector
        y : arralike
            y vector
        z : arraylike
            z vector
        V : arraylike
            3D array with data corresponding to vectors X x Y x Z direction.
        labeldict : dict, optional
            dictionary with labels, by default {'moho': 1, '410': 2, '660': 3, 'none': 9999}
        labeled : Optional[dict], optional
            Optional dictionary with labeled volume. Labeled should contain 
            4 variables lx, ly, lz, lV, where l{x,y,z} are boolean arrays that
            where defined as locations for labeled layers. 
            Useful if you need to take a labeling break. Right now there is no 
            way of resuming to an image make sure you finisshe the ones you
            have started, by default None

        Notes
        -----

        Things that I have to fix:

        1. The volume should be allowed to be 3-channel data
        2. Also allow for custom normalizations and colormaps

        :Authors:
            Lucas Sawade (lsawade@princeton.edu)

        :Last Modified:
            2021.07.02 00.00 (Lucas Sawade)

        """

        # Variables
        self.x = x
        self.y = y
        self.z = z
        self.V = V

        # Dimensions
        self.nx = len(x)
        self.ny = len(y)
        self.nz = len(z)

        # Label info
        self.labeldict = labeldict

        # Load labeled volume, to continue labeling
        if labeled is not None:
            self.labeled = labeled
        else:
            self.labeled = dict()
            self.labeled['lx'] = np.zeros(self.nx, dtype=bool)
            self.labeled['ly'] = np.zeros(self.ny, dtype=bool)
            self.labeled['lz'] = np.zeros(self.ny, dtype=bool)
            self.labeled['lV'] = self.labeldict['none'] * \
                np.ones(self.V.shape, dtype=int)

        # Colormap choices
        self.cmap = plt.get_cmap('seismic')
        self.norm = MidpointNormalize(midpoint=0, vmin=np.min(
            self.V), vmax=0.25 * np.max(self.V))

    @classmethod
    def from_volume(self, filename, *args, **kwargs):

        # Load NPZ file

        vardict = np.load(filename)

        # Assign variables
        x = vardict["x"]
        y = vardict["y"]
        z = vardict["z"]
        V = vardict["data"]

        return self(x, y, z, V, *args, **kwargs)

    @classmethod
    def from_labeled_volume(self, filename, *args, **kwargs):

        # Load NPZ file
        vardict = np.load(filename, allow_pickle=True)

        # Assign variables
        x = vardict["x"]
        y = vardict["y"]
        z = vardict["z"]
        V = vardict["V"]

        # Labeled
        labeled = dict()
        labeled["lx"] = vardict["lx"]
        labeled["ly"] = vardict["ly"]
        labeled["lz"] = vardict["lz"]
        labeled["lV"] = vardict["lV"]

        # Label dictionary
        labeldict = vardict["labeldict"]

        return self(
            x, y, z, V, *args,
            labeled=labeled, labeldict=labeldict, **kwargs)

    def save(self, outfile):

        np.savez(
            outfile,
            x=self.x, y=self.y, z=self.z, V=self.V,
            lx=self.labeled['lx'],
            ly=self.labeled['ly'],
            lz=self.labeled['lz'],
            lV=self.labeled['lV'],
            labeldict=self.labeldict
        )

    def label(self, direction='x', n=20, n_segments=700):
        """The labeling method enables you to label a ``n`` slices
        The labeler will automatically pick ``n`` unlabeled
        slices randomly. The direction let's you decide the normal to the 
        slices.

        Parameters
        ----------
        direction : str, optional
            normal to slices, by default 'x'
        n : int, optional
            number of slices tto label, by default 20
        n_segments : int, optional
            number of segments to split the image into, by default 700

        Raises
        ------
        ValueError
            [description]
        """

        # Get slice to plot
        if direction == 'x':
            tobelabeled = self.__get_randarray__(self.labeled["lx"], n)
        elif direction == 'y':
            tobelabeled = self.__get_randarray__(self.labeled["ly"], n)
        elif direction == 'z':
            tobelabeled = self.__get_randarray__(self.labeled["lz"], n)
        else:
            raise ValueError(f'Direction {direction} not implemented.')

        for _label_idx in tobelabeled:
            try:
                # Get slice to plot
                if direction == 'x':
                    array = self.V[_label_idx, :, :].T
                elif direction == 'y':
                    array = self.V[:, _label_idx, :].T
                elif direction == 'z':
                    array = self.V[:, :, _label_idx]

                img2 = self.cmap(self.norm(array))
                img = img2[:, :, :-1]

                # Super Pixel segmentation
                segments = slic(
                    img, n_segments=n_segments, compactness=20, sigma=0,
                    start_label=0)

                # Label
                sl = SegmentLabeler(img, segments, labeldict=self.labeldict)

                # Get slice to plot
                if direction == 'x':
                    self.labeled["lV"][_label_idx,
                                       :, :] = sl.start_labeling().T
                    self.labeled["lx"][_label_idx] = True
                elif direction == 'y':
                    self.labeled["lV"][:, _label_idx,
                                       :] = sl.start_labeling().T
                    self.labeled["ly"][_label_idx] = True
                elif direction == 'z':
                    self.labeled["lV"][:, :, _label_idx] = sl.start_labeling()
                    self.labeled["lz"][_label_idx] = True
            except KeyboardInterrupt:
                break

    @staticmethod
    def __get_randarray__(labeled, N):

        # Get Unlabeled indeces
        unlabeled_indeces = np.where(~labeled)[0]

        # Check whether all labeled
        if len(unlabeled_indeces) == 0:
            raise ValueError('Nothing to be labeled')
        elif len(unlabeled_indeces) < N:
            print('Less to be labeled than indicated.')
            N = len(unlabeled_indeces)

        # Choose N times from unlabeled indeces
        tobelabeled = np.random.choice(
            unlabeled_indeces, size=N, replace=False)

        return tobelabeled


if __name__ == "__main__":

    infile = '/Users/lucassawade/OneDrive/Research/RF/DATA/GLImER/ccps/US_P_0.58_minrad_3D_it_f2_volume.pkl.npz'

    outfile = 'labeled_volume.npz'

    S = Segmenter.from_CCPVolume(infile)

    S.label(direction='y', n=20, n_segments=700)

    S.save(outfile)
